//$ bin/sjc tests/src/ticketagency/Customer.sj -d tests/classes/
//$ bin/sj -cp tests/classes/ ticketagency.Customer localhost 9999 100.00

package ticketagency;

import java.net.UnknownHostException;
import java.util.Date;

import sj.runtime.*;
import sj.runtime.net.*;

class Customer 
{
	private static protocol p_ca 
	{
		begin
		.![
			!<String>
			.?(Double)
		]*
		.!{
				ACCEPT: !<Address>.?(Date)
			, REJECT:
		}
	}

	private String TRAVEL_METHOD = "Paris by Eurostar";
	private Double MAX_PRICE;
	private Address ADDRESS = new Address("Customer's address.");

	Customer(String addr_a, int port_a, double maxPrice) throws SJIOException
	{
		MAX_PRICE = new Double(maxPrice);

		SJServerAddress c_ca = SJServerAddress.create(p_ca, addr_a, port_a);
		SJSocket s_ca = SJRSocket.create(c_ca);
		
		boolean decided = false;
		int retry = 3;

		try (s_ca)
		{
			s_ca.request();
			
			s_ca.outwhile(!decided && (retry-- > 0)) 
			{
				s_ca.send(TRAVEL_METHOD);
				
				Double cost = s_ca.receive();
				
				System.out.println("Received quote: " + cost);
				
				if (cost.compareTo(MAX_PRICE) < 0) 
				{
					decided = true;
				}
			}
			if (retry >= 0) 
			{
				s_ca.outbranch(ACCEPT) 
				{
					
					System.out.println("Quote accepted.");
					
					s_ca.send(ADDRESS);
					
					//System.out.println("Dispatch date: " + (Date)s_ca.receive());
					System.out.println("Received dispatch date: " + s_ca.receive());

				}
			}
			else 
			{
				s_ca.outbranch(REJECT) 
				{
				
					System.out.println("Quote rejected.");
					
				}			
			}
		}
		catch (SJIncompatibleSessionException ise) 
		{		
			ise.printStackTrace();
		}
		catch (UnknownHostException uhe) 
		{ 
			uhe.printStackTrace(); 
		}
		catch (ClassNotFoundException cnfe) { }
	}

	public static void main(String[] args) throws SJIOException
	{
		new Customer(args[0], Integer.parseInt(args[1]), Double.parseDouble(args[2]));
	}
}
